/*
Walk the MCB List and print it out.
by Bloody Cactus
*/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <inttypes.h>
#include <assert.h>
#include <string.h>
#include <i86.h>
#include "dlist.h"


#define MEM_ID_OK	0x4D
#define MEM_ID_LAST	0x5A

#define SEG_FREE	0
#define SEG_SYSTEM	0x08

#pragma pack(1);

typedef struct udtMCB
{
	uint8_t magic;
	uint16_t seg_owner;
	uint16_t size_paras;
	uint8_t reserved[3];
	uint8_t name[8];
} uMCB;

typedef struct udtMemList
{
	uint8_t magic;
	uint16_t seg_owner;
	uint16_t size_paras;
	uint16_t seg;
	uint8_t name[9];
} uMemList;

#pragma pack()

typedef struct udtAddrRange
{
	uint16_t start;
	uint16_t end;
	uint8_t done;
} uAddrRange;


extern uint16_t dos_list_of_lists(void);

#pragma aux dos_list_of_lists = \
    "mov ax,0x5200h"       \
    "int 21h"            \
    "mov ax,[es: bx-2]" \
    value     [ax]     \
    modify    [es bx];

extern uint16_t dos_version(void);

#pragma aux dos_version = \
    "mov ax,0x3000h"       \
    "int 21h"            \
    "xchg al,ah" \
    value     [ax]     \
    modify    [bx cx];

extern uint16_t dos_add_umbs(void);
#pragma aux dos_add_umbs = \
    "mov ax,0x5803h"       \
    "mov bx,1" \
    "int 21h"            \
    modify    [ax bx cx];

extern uint16_t dos_remove_umbs(void);
#pragma aux dos_remove_umbs = \
    "mov ax,0x5803h"       \
    "xor bx,bx" \
    "int 21h"            \
    modify    [ax bx cx];

/*****************************************************************************/

char *own_name;

static DList* BuildMCBList(void)
{
	uint16_t seg;
	uMCB far *x;
	uMemList far *copy;
	DList *lst;
	char far *owner_name;
	int done = 0;

	if(dos_version() >= 0x500)
	{
		dos_add_umbs();
	}

	lst = NewDList(free);
	seg = dos_list_of_lists();

	x = (uMCB far*)MK_FP(seg, 0);

	do
	{
		if(x->magic == MEM_ID_LAST)
		{
			done = 1;
		}

		copy = calloc(1, sizeof(uMemList));

		copy->magic = x->magic;
		copy->size_paras = x->size_paras;
		copy->seg_owner = x->seg_owner;
		copy->seg = seg;

		if(x->seg_owner != SEG_FREE && x->seg_owner != SEG_SYSTEM)
		{
			int i;

			owner_name = MK_FP(x->seg_owner - 1, 8);
			if(*owner_name != 0)
			{
				memmove(copy->name, owner_name, 8);
			}

			// dos 3 does not put name in block?
			for(i=0; i < 8; i++)
			{
				if(copy->name[i] !=0 && (copy->name[i] < 0x20 || copy->name[i] > 0x7F))
					memset(copy->name, 0x0, 8);
			}
		}

		dlist_ins(lst, copy);

		seg = seg + x->size_paras + 1;
		x = (uMCB far*)MK_FP(seg, 0);
	}while(done == 0);


	if(dos_version() >= 0x500)
	{
		dos_remove_umbs();
	}


	return lst;
}

static void ListChain(void)
{
	DList *lstSegs;
	DLElement *e;
	uMemList *mem;
	uMemList *work;
	uint16_t kbs;
	uint16_t total_mem;
	uint16_t free_mem;
	uint16_t larget_free_mem;
	uint16_t largest_free_umb;

	total_mem = 0;
	free_mem = 0;
	larget_free_mem = 0;
	largest_free_umb = 0;

	lstSegs = BuildMCBList();

	// remove ourself from the chain and conflate empty entries above and below
	e = dlist_head(lstSegs);
	while(e != NULL)
	{
		mem = dlist_data(e);

		if(stricmp(mem->name, own_name) == 0)
		{
			// change from US to FREE
			mem->seg_owner = SEG_FREE;
			mem->name[0] = 0;

			// test previous
			if(dlist_head(lstSegs) != e)
			{
				work = dlist_data(dlist_prev(e));
				if(work->seg_owner == SEG_FREE)
				{
					if(work->seg + work->size_paras + 1 == mem->seg)
					{
						// conflate with previous
						work->size_paras += 1 + mem->size_paras;
						e = dlist_prev(e);
						dlist_remove(lstSegs, dlist_next(e), (void*)&work);

						mem = dlist_data(e);
					}
				}
			}

			// test next
			if(dlist_tail(lstSegs) != e)
			{
				// conflate above
				work = dlist_data(dlist_next(e));
				if(work->seg_owner == SEG_FREE)
				{
					if(mem->size_paras + 1 + mem->seg == work->seg)
					{
						// conflate with previous
						mem->size_paras += 1 + work->size_paras;
						// copy magic from above incase its last in chain!
						mem->magic = work->magic;
						dlist_remove(lstSegs, dlist_next(e), (void*)&work);
					}
				}
			}
		}

		e = dlist_next(e);
	}

	printf(" Segment     Total    Notes\n");
	printf("  -----    -------    -----------------------------\n");

	e = dlist_head(lstSegs);
	while(e != NULL)
	{
		mem = dlist_data(e);
		e = dlist_next(e);

		kbs = 1 + mem->size_paras;

		printf("  %05X    %5ukb    ", mem->seg, kbs / 64);
		if(mem->seg_owner == SEG_FREE) // || strcmp(mem->name, own_name) == 0)
			printf("-- Free --\n");
		else if(mem->seg_owner == SEG_SYSTEM)
			printf("System\n");
		else if(mem->name[0] == 0)
			printf("Used\n");
		else
			printf("%s\n", mem->name);


		total_mem += mem->size_paras + 1;
		if(mem->seg_owner == SEG_FREE)
		{
			free_mem += mem->size_paras + 1;
			if(mem->seg < 0xA000)
			{
				if(mem->size_paras + 1 > larget_free_mem)
				{
					larget_free_mem = mem->size_paras + 1;					
				}
			}
			else
			{
				if(mem->size_paras + 1 > largest_free_umb)
				{
					largest_free_umb = mem->size_paras + 1;
				}
			}
		}
	}

	printf("\n");
	printf("Total Memory : %ukb\n", total_mem / 64);
	printf("Total Free Memory : %ukb\n", free_mem / 64);
	printf("Largest free low memory size : %ukb\n", larget_free_mem / 64);
	printf("Largest free UMB size : %ukb\n", largest_free_umb / 64);
}

int main(int argc, char *argv[])
{
	char *p;

	p = strchr(argv[0], 0x0);
	while(p != argv[0] && *p != '/' && *p != '\\')
		p--;

	if(*p == '/' || *p == '\\')
		own_name = strdup(p+1);
	else
		own_name = strdup(p);

	p = strchr(own_name, '.');
	if(p != NULL)
		*p = 0;

	strupr(own_name);

	ListChain();

	return 0;
}
